﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Delegates {

	delegate void Proc(int a);

	class A {
		public int m;

		public void f(int x) {
			Console.WriteLine("instance A.f(x == {0}): m == {1}", x, m);
		}

		public static void g(int x) {
			Console.WriteLine("static A.g(x == {0})", x);
		}

		private static void privateH(int x) {
			Console.WriteLine("private static A.g(x == {0})", x);
		}

		public static Proc UsefulMethod => new Proc(privateH);
	}

	class Program {
		static void Main(string[] args) {
			A a1 = new A { m = 1 };
			A a2 = new A { m = 2 };

			Proc p;
			p = new Proc(A.g);

			p.Invoke(5);

			p(10);

			p = A.UsefulMethod;
			p(11);

			p = new Proc(a1.f);
			p(20);

			p = a2.f;
			p(25);

			p = null;
			try {
				p(30);
			} catch (Exception ex) {
				Console.WriteLine(ex);
			}

			Console.WriteLine();
			Console.Write("Press ENTER to get to more complex examples ...");
			Console.ReadLine();

			p = null;
			// p = (Proc) Delegate.Combine(p, a1.f);
			p = (Proc) Delegate.Combine(p, new Proc(a1.f));
			p(40); Console.WriteLine();
			
			p = (Proc) Delegate.Combine(p, new Proc(a2.f));
			p(50); Console.WriteLine();
			
			p = (Proc) Delegate.Combine(p, new Proc(a1.f));
			p(60); Console.WriteLine();

			Delegate[] list = p.GetInvocationList();
			for (int i = 0; i < list.Length; i++) {
				Console.Write($"Calling delegate #{i} of {list[i].Method} with A.m == {((A) list[i].Target).m}: ");
				((Proc) list[i])(65);
			}
			Console.WriteLine();

			p = (Proc) Delegate.Remove(p, new Proc(a1.f));
			p(70); Console.WriteLine();

			Console.Write("Press ENTER to get to more complex examples ...");
			Console.ReadLine();

			p = null;
			try {
				p(80);
			} catch (Exception ex) {
				Console.WriteLine(ex);
			}

			p += a1.f;
			p(90); Console.WriteLine();

			p += a2.f;
			p(100); Console.WriteLine();

			p -= new Proc(a1.f);
			p(200); Console.WriteLine();

			p -= new Proc(a2.f);
			try {
				p(200);
			} catch (Exception ex) {
				Console.WriteLine(ex);
			}
		}
	}
}
